#version 300 es

precision highp float;
precision highp sampler3D;

const vec3 satLuminanceWeighting = vec3(0.2126, 0.7152, 0.0722);
const vec3 hsLuminanceWeighting = vec3(0.3, 0.3, 0.3);

uniform sampler2D u_Texture;
uniform float u_Contrast;
uniform float u_Brightness;
uniform float u_Saturation;
uniform float u_Shadows;
uniform float u_Highlights;
uniform float u_ShadowsTintIntensity;
uniform vec3 u_ShadowsTintColor;
uniform float u_FadeAmount;
uniform float u_Warmth;
uniform float u_Exposure;
uniform float u_Sharpen;
uniform float u_Grain;
uniform float u_SkinTone;
uniform float u_Opacity;

uniform vec4 u_Blend;
uniform int u_Mode;

uniform int u_ApplyCurves;
uniform sampler2D u_CurvesImage;

uniform int u_ApplyLookUp;
uniform sampler2D u_LookUp;
uniform vec2 u_LookUpSize;
uniform float u_LookUpColors;

in vec2 v_LeftTexCoord;
in vec2 v_RightTexCoord;
in vec2 v_TopTexCoord;
in vec2 v_BottomTexCoord;
in vec2 v_TexCoord;

layout(location = 0) out vec4 o_FragColor;

const float SKIN_HUE = 0.05;
const float SKIN_HUE_TOLERANCE = 50.0;
const float MAX_HUE_SHIFT = 0.04;
const float MAX_SATURATION_SHIFT = 0.25;

vec4 lookup(vec4 textureColor) {
    float maxColor = u_LookUpColors - 1.0;
    float halfColX = 0.5 / u_LookUpSize.x;
    float halfColY = 0.5 / u_LookUpSize.y;
    float threshold = maxColor / u_LookUpColors;

    float xOffset = halfColX + textureColor.r * threshold / u_LookUpColors;
    float yOffset = halfColY + textureColor.g * threshold;
    float cell = floor(textureColor.b * maxColor);

    vec2 lutPos = vec2(cell / u_LookUpColors + xOffset, yOffset);

    return texture(u_LookUp, lutPos);
}


vec3 rgb2hsv(vec3 c)
{
    vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
    vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
    vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

    float d = q.x - min(q.w, q.y);
    float e = 1.0e-10;
    return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}

vec3 hsv2rgb(vec3 c)
{
    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

 vec3 rgbToHsl( vec3 color) {
    vec3 hsl;
    float fmin = min(min(color.r, color.g), color.b);
    float fmax = max(max(color.r, color.g), color.b);
    float delta = fmax - fmin;
    hsl.z = (fmax + fmin) / 2.0;

    if (delta == 0.0) {
        hsl.x = 0.0;
        hsl.y = 0.0;
    } else {
        if (hsl.z < 0.5) {
            hsl.y = delta / (fmax + fmin);
        } else {
            hsl.y = delta / (2.0 - fmax - fmin);
        }

        float deltaR = (((fmax - color.r) / 6.0) + (delta / 2.0)) / delta;
        float deltaG = (((fmax - color.g) / 6.0) + (delta / 2.0)) / delta;
        float deltaB = (((fmax - color.b) / 6.0) + (delta / 2.0)) / delta;

        if (color.r == fmax) {
            hsl.x = deltaB - deltaG;
        } else if (color.g == fmax) {
            hsl.x = (1.0 / 3.0) + deltaR - deltaB;
        } else if (color.b == fmax) {
            hsl.x = (2.0 / 3.0) + deltaG - deltaR;
        }

        if (hsl.x < 0.0) {
            hsl.x += 1.0;
        } else if (hsl.x > 1.0) {
            hsl.x -= 1.0;
        }
    }
    return hsl;
}
 float hueToRgb(float f1, float f2, float hue) {
    if (hue < 0.0) {
        hue += 1.0;
    } else if (hue > 1.0) {
        hue -= 1.0;
    }

     float res;


    if ((6.0 * hue) < 1.0) {
        res = f1 + (f2 - f1) * 6.0 * hue;
    } else if ((2.0 * hue) < 1.0) {
        res = f2;
    } else if ((3.0 * hue) < 2.0) {
        res = f1 + (f2 - f1) * ((2.0 / 3.0) - hue) * 6.0;
    } else {
        res = f1;
    } return res;
}

vec3 hslToRgb( vec3 hsl) {
    if (hsl.y == 0.0) {
        return vec3(hsl.z);
    } else {
        float f2;
        if (hsl.z < 0.5) {
            f2 = hsl.z * (1.0 + hsl.y);
        } else {
            f2 = (hsl.z + hsl.y) - (hsl.y * hsl.z);
        }

        float f1 = 2.0 * hsl.z - f2;
        return vec3(hueToRgb(f1, f2, hsl.x + (1.0/3.0)), hueToRgb(f1, f2, hsl.x), hueToRgb(f1, f2, hsl.x - (1.0/3.0)));
    }
}

vec3 applyLuminanceCurve( vec3 pixel) {
     float index = floor(clamp(pixel.z / (1.0 / 200.0), 0.0, 199.0));
    pixel.y = mix(0.0, pixel.y, smoothstep(0.0, 0.1, pixel.z) * (1.0 - smoothstep(0.8, 1.0, pixel.z)));
    pixel.z = texture(u_CurvesImage, vec2(1.0 / 200.0 * index, 0)).a;
    return pixel;
}

vec3 applyRGBCurve(vec3 pixel) {
     float index = floor(clamp(pixel.r / (1.0 / 200.0), 0.0, 199.0));
    pixel.r = texture(u_CurvesImage, vec2(1.0 / 200.0 * index, 0)).r;
    index = floor(clamp(pixel.g / (1.0 / 200.0), 0.0, 199.0));
    pixel.g = clamp(texture(u_CurvesImage, vec2(1.0 / 200.0 * index, 0)).g, 0.0, 1.0);
    index = floor(clamp(pixel.b / (1.0 / 200.0), 0.0, 199.0));
    pixel.b = clamp(texture(u_CurvesImage, vec2(1.0 / 200.0 * index, 0)).b, 0.0, 1.0);
    return pixel;
}

vec3 tintRaiseShadowsCurve(vec3 color) {
    return vec3(-0.003671) * pow(color, vec3(3.0)) + vec3(0.3842) * pow(color, vec3(2.0)) + vec3(0.3764) * color + vec3(0.2515);
}

vec3 tintShadows(vec3 texel, vec3 tintColor, float tintAmount) {
    return clamp(mix(texel, mix(texel, tintRaiseShadowsCurve(texel), tintColor), tintAmount), 0.0, 1.0);
}

vec3 fadeAdjust(vec3 color, float fadeVal) {
    return (color * (1.0 - fadeVal)) + ((color + (vec3(-0.9772) * pow(vec3(color), vec3(3.0)) + vec3(1.708) * pow(vec3(color), vec3(2.0)) + vec3(-0.1603) * vec3(color) + vec3(0.2878) - color * vec3(0.9))) * fadeVal);
}

float getLuma(vec3 rgbP) {
    return (0.299 * rgbP.r) + (0.587 * rgbP.g) + (0.114 * rgbP.b);
}

vec3 rgbToYuv(vec3 inP) {
    float luma = getLuma(inP);

    return vec3(luma, (1.0 / 1.772) * (inP.b - luma), (1.0 / 1.402) * (inP.r - luma));
}

vec3 yuvToRgb(vec3 inP) {
    return vec3(1.402 * inP.b + inP.r, (inP.r - (0.299 * 1.402 / 0.587) * inP.b - (0.114 * 1.772 / 0.587) * inP.g), 1.772 * inP.g + inP.r);
}

float blendOverlay(float base, float blend) {
    return base<0.5?(2.0*base*blend):(1.0-2.0*(1.0-base)*(1.0-blend));
}

vec3 blendOverlay(vec3 base, vec3 blend) {
    return vec3(blendOverlay(base.r,blend.r),blendOverlay(base.g,blend.g),blendOverlay(base.b,blend.b));
}

vec3 blendOverlay(vec3 base, vec3 blend, float opacity) {
    return (blendOverlay(base, blend) * opacity + base * (1.0 - opacity));
}

vec3 blendMultiply(vec3 base, vec3 blend) {
    return base*blend;
}

vec3 blendMultiply(vec3 base, vec3 blend, float opacity) {
    return (blendMultiply(base, blend) * opacity + base * (1.0 - opacity));
}

float blendScreen(float base, float blend) {
    return 1.0-((1.0-base)*(1.0-blend));
}

vec3 blendScreen(vec3 base, vec3 blend) {
    return vec3(blendScreen(base.r,blend.r),blendScreen(base.g,blend.g),blendScreen(base.b,blend.b));
}

vec3 blendScreen(vec3 base, vec3 blend, float opacity) {
    return (blendScreen(base, blend) * opacity + base * (1.0 - opacity));
}

float blendSoftLight(float base, float blend) {
    return (blend<0.5)?(2.0*base*blend+base*base*(1.0-2.0*blend)):(sqrt(base)*(2.0*blend-1.0)+2.0*base*(1.0-blend));
}

vec3 blendSoftLight(vec3 base, vec3 blend) {
    return vec3(blendSoftLight(base.r,blend.r),blendSoftLight(base.g,blend.g),blendSoftLight(base.b,blend.b));
}

vec3 blendSoftLight(vec3 base, vec3 blend, float opacity) {
    return (blendSoftLight(base, blend) * opacity + base * (1.0 - opacity));
}

float blendDarken(float base, float blend) {
    return min(blend,base);
}

vec3 blendDarken(vec3 base, vec3 blend) {
    return vec3(blendDarken(base.r,blend.r),blendDarken(base.g,blend.g),blendDarken(base.b,blend.b));
}

vec3 blendDarken(vec3 base, vec3 blend, float opacity) {
    return (blendDarken(base, blend) * opacity + base * (1.0 - opacity));
}

float blendLighten(float base, float blend) {
    return max(blend,base);
}

vec3 blendLighten(vec3 base, vec3 blend) {
    return vec3(blendLighten(base.r,blend.r),blendLighten(base.g,blend.g),blendLighten(base.b,blend.b));
}

vec3 blendLighten(vec3 base, vec3 blend, float opacity) {
    return (blendLighten(base, blend) * opacity + base * (1.0 - opacity));
}

vec3 blendHardLight(vec3 base, vec3 blend) {
    return blendOverlay(blend,base);
}

vec3 blendHardLight(vec3 base, vec3 blend, float opacity) {
    return (blendHardLight(base, blend) * opacity + base * (1.0 - opacity));
}

float blendPinLight(float base, float blend) {
    return (blend<0.5)?blendDarken(base,(2.0*blend)):blendLighten(base,(2.0*(blend-0.5)));
}

vec3 blendPinLight(vec3 base, vec3 blend) {
    return vec3(blendPinLight(base.r,blend.r),blendPinLight(base.g,blend.g),blendPinLight(base.b,blend.b));
}

vec3 blendPinLight(vec3 base, vec3 blend, float opacity) {
    return (blendPinLight(base, blend) * opacity + base * (1.0 - opacity));
}

vec3 blendMode(vec3 base, vec3 blend, float a){
    if (u_Mode == 0){
        return base * (1.0 - a) + blend;
    } else if (u_Mode == 2) {
        return blendMultiply(base, blend, a);
    } if (u_Mode == 4) {
        return blendScreen(base, blend, a);
    } else if (u_Mode == 5) {
        return blendOverlay(base, blend, a);
    } else if (u_Mode == 6) {
        return blendSoftLight(base, blend, a);
    } else if (u_Mode == 1) {
        return blendDarken(base, blend, a);
    } else if (u_Mode == 3) {
        return blendLighten(base, blend, a);
    } else if (u_Mode == 7) {
        return blendHardLight(base, blend, a);
    } else if (u_Mode == 8) {
        return blendPinLight(base, blend, a);
    } else {
        return blend;
    }
}

void main(void) {
    const float toolEpsilon = 0.005;
    vec4 source = texture(u_Texture, v_TexCoord);
    vec4 result = source;

    if (u_ApplyLookUp == 1) {
        result = lookup(source);
    } else {
        if (u_ApplyCurves == 1) {
            result = vec4(applyRGBCurve(hslToRgb(applyLuminanceCurve(rgbToHsl(result.rgb / result.a)))), 1.0) * result.a;
        }

        float hsLuminance = dot(result.rgb, hsLuminanceWeighting);
        float shadow = clamp((pow(hsLuminance, 1.0 / u_Shadows) + (-0.76) * pow(hsLuminance, 2.0 / u_Shadows)) - hsLuminance, 0.0, 1.0);
        float highlight = clamp((1.0 - (pow(1.0 - hsLuminance, 1.0 / (2.0 - u_Highlights)) + (-0.8) * pow(1.0 - hsLuminance, 2.0 / (2.0 - u_Highlights)))) - hsLuminance, -1.0, 0.0);
        vec3 hsresult = vec3(0.0, 0.0, 0.0) + ((hsLuminance + shadow + highlight) - 0.0) * ((result.rgb - vec3(0.0, 0.0, 0.0)) / (hsLuminance - 0.0));
        float contrastedLuminance = ((hsLuminance - 0.5) * 1.5) + 0.5;
        float whiteInterp = contrastedLuminance * contrastedLuminance * contrastedLuminance;
        float whiteTarget = clamp(u_Highlights, 1.0, 2.0) - 1.0;
        hsresult = mix(hsresult, vec3(1.0), whiteInterp * whiteTarget);
        float invContrastedLuminance = 1.0 - contrastedLuminance;
        float blackInterp = invContrastedLuminance * invContrastedLuminance * invContrastedLuminance;
        float blackTarget = 1.0 - clamp(u_Shadows, 0.0, 1.0);
        hsresult = mix(hsresult, vec3(0.0), blackInterp * blackTarget);
        result = vec4(hsresult.rgb, result.a);

        result = vec4(clamp(((result.rgb - vec3(0.5)) + u_Brightness + vec3(0.5)), 0.0, 1.0), result.a);

        result = vec4(clamp(((result.rgb / result.a - vec3(0.5)) * u_Contrast + vec3(0.5)), 0.0, 1.0), 1.0) * result.a;

        if (abs(u_FadeAmount) > toolEpsilon) {
            result.rgb = fadeAdjust(result.rgb / result.a, u_FadeAmount) * result.a;
        }

        float satLuminance = dot(result.rgb, satLuminanceWeighting);
        vec3 greyScaleColor = vec3(satLuminance);
        result = vec4(clamp(mix(greyScaleColor, result.rgb, u_Saturation), 0.0, 1.0), result.a);

        if (abs(u_ShadowsTintIntensity)>toolEpsilon) {
            result.rgb = tintShadows(result.rgb, u_ShadowsTintColor, u_ShadowsTintIntensity * 2.0);
        }

        if (abs(u_Exposure)>toolEpsilon) {
            float mag = u_Exposure * 1.045;
            float exppower = 1.0 + abs(mag);

            if (mag < 0.0) {
                exppower = 1.0 / exppower;
            }

            result.r = 1.0 - pow((1.0 - result.r), exppower);
            result.g = 1.0 - pow((1.0 - result.g), exppower);
            result.b = 1.0 - pow((1.0 - result.b), exppower);
        }

        if (abs(u_Warmth)>toolEpsilon){
            vec3 yuvVec;

            if (u_Warmth > 0.0) {
                yuvVec = vec3(0.1765, -0.1255, 0.0902);
            } else {
                yuvVec = -vec3(0.0588, 0.1569, -0.1255);
            }

            vec3 yuvColor = rgbToYuv(result.rgb);
            float luma = yuvColor.r;
            float curveScale = sin(luma * 3.14159);
            yuvColor += 0.375 * u_Warmth * curveScale * yuvVec;
            result.rgb = yuvToRgb(yuvColor);
        }

        vec3 leftTextureColor = texture(u_Texture, v_LeftTexCoord).rgb;
        vec3 rightTextureColor = texture(u_Texture, v_RightTexCoord).rgb;
        vec3 topTextureColor = texture(u_Texture, v_TopTexCoord).rgb;
        vec3 bottomTextureColor = texture(u_Texture, v_BottomTexCoord).rgb;
        result.rgb = result.rgb * (1.0 + 4.0 * u_Sharpen) - (leftTextureColor + rightTextureColor + topTextureColor + bottomTextureColor) * u_Sharpen;

        if (abs(u_SkinTone - 0.5) > toolEpsilon) {
            // mouse Y => skin_tone_shift in [-1, 1]
            float skin_tone_shift = (u_SkinTone * 2.0 - 1.0);
            //skin_tone_shift = clamp(skin_tone_shift * 1.25, -1.0, 1.0);

            // Convert color to HSV, extract hue
            vec3 colorHSV = rgb2hsv(result.rgb);
            float hue = colorHSV.x;

            // check how far from skin hue
            float dist = hue - SKIN_HUE;
            if (dist > 0.5) {
                dist -= 1.0;
            }

            if (dist < -0.5) {
                dist += 1.0;
            }

            dist = abs(dist)/0.5;// normalized to [0,1]

            // Apply Gaussian like filter
            float weight = exp(-dist*dist*SKIN_HUE_TOLERANCE);
            weight = clamp(weight, 0.0, 1.0);

            if (skin_tone_shift > 0.0) {
                // We want more orange, so increase saturation
                colorHSV.y += skin_tone_shift * weight * MAX_SATURATION_SHIFT;
            } else {
                // we want more pinks, so decrease hue
                colorHSV.x += skin_tone_shift * weight * MAX_HUE_SHIFT;
            }

            // final color
            result.rgb = hsv2rgb(colorHSV.rgb);
        }

        if (u_Grain > toolEpsilon) {
            vec2 uv = v_TexCoord;
            float strength = 64.0 * u_Grain;
            float x = (uv.x + 4.0) * (uv.y + 4.0) * (/*iTime*/ 1.0 * 10.0);
            vec3 grain = vec3(mod((mod(x, 13.0) + 1.0) * (mod(x, 123.0) + 1.0), 0.01)-0.005) * strength;

            //        if(uv.x > 0.5)
            //        {
            grain = 1.0 - grain;
            result.rgb = result.rgb * grain;
            //        }
            //        else
            //        {
            //            result.rgb = result.rgb + grain;
            //        }
        }

        if (u_Mode != -1 && u_Blend.a > 0.0) {
            result = vec4(blendMode(result.rgb, u_Blend.rgb, u_Blend.a), result.a);
        }
    }

    o_FragColor = mix(source, result, u_Opacity);
}